package tools.descartes.dql.connector.kieker;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EOperation;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.util.EcoreUtil;

import kieker.common.record.system.CPUUtilizationRecord;
import tools.descartes.dql.connector.AbstractOSGiQueryConnector;
import tools.descartes.dql.connector.PerformanceMetricsQueryConnector;
import tools.descartes.dql.connector.QueryProgressListener;
import tools.descartes.dql.connector.kieker.filter.SessionFilter;
import tools.descartes.dql.connector.kieker.structures.Component;
import tools.descartes.dql.connector.kieker.structures.Resource;
import tools.descartes.dql.connector.kieker.structures.ResourceContent;
import tools.descartes.dql.connector.kieker.structures.ResourceContent.ResourceContentType;
import tools.descartes.dql.connector.kieker.structures.Session;
import tools.descartes.dql.connector.kieker.structures.Trace;
import tools.descartes.dql.connector.kieker.structures.identifier.ResourceIdentifier;
import tools.descartes.dql.connector.kieker.structures.records.EventRecord;
import tools.descartes.dql.connector.kieker.structures.records.ResourceRecord;
import tools.descartes.dql.connector.kieker.util.KiekerHelper;
import tools.descartes.dql.connector.kieker.util.KiekerManager;
import tools.descartes.dql.core.engine.util.DQLLogger;
import tools.descartes.dql.core.engine.util.DoFCrossProductHelper;
import tools.descartes.dql.models.mapping.domain.ContinuousSeriesResult;
import tools.descartes.dql.models.mapping.domain.DecimalResult;
import tools.descartes.dql.models.mapping.domain.DoF;
import tools.descartes.dql.models.mapping.domain.DomainFactory;
import tools.descartes.dql.models.mapping.domain.Entity;
import tools.descartes.dql.models.mapping.domain.Probe;
import tools.descartes.dql.models.mapping.domain.SeriesResultElement;
import tools.descartes.dql.models.mapping.domain.StatType;
import tools.descartes.dql.models.mapping.domain.StringResult;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;
import tools.descartes.dql.models.mapping.mapping.MappingFactory;

public class PerformanceMetricsQueryConnectorImpl extends AbstractOSGiQueryConnector
		implements PerformanceMetricsQueryConnector {

	private final DomainFactory domainFactory = DomainFactory.eINSTANCE;
	
	private static final Logger log = DQLLogger.getLogger(PerformanceMetricsQueryConnectorImpl.class.getName());


	@Override
	public EntityMapping requestMetrics(final EntityMapping request) {

		final EntityMapping response = EcoreUtil.copy(request);

		KiekerManager manager = new KiekerManager(request);
		manager.run();
		Trace result = manager.getAnalysisResult();
		long firstTs = manager.getStartingTime();
		long lastTs = manager.getEndTime();

		// Process the trace back to a DQL result
		// Kieker ergebnisse filtern
		processPerformanceResults(response, result, firstTs, lastTs);
		return response;
	}
	
	@Override
	public List<EntityMapping> requestDoFMetrics(final EntityMapping request) {
		final ArrayList<EntityMapping> results = new ArrayList<EntityMapping>();
			
		String location = request.getModelLocation();
		int index = location.lastIndexOf(File.separator);
		String substring = location.substring(index);
		boolean isPropertiesFileVariation;
		if(substring.contains(".properties")) {
			log.info("Varying properties configuration");
			isPropertiesFileVariation = true;
			substring = substring.replace(".properties", "");
			substring.replaceAll("\\d","");
		}else{
			log.info("Varying path to search subdirectories");			
			isPropertiesFileVariation = false;
		}
		
		List<HashMap<DoF, String>> doFs = new DoFCrossProductHelper().calculateCrossProduct(request);
		for (HashMap<DoF, String> configuration : doFs) {
			for (Entry<DoF, String> configurationPoint : configuration.entrySet()) {
				if(isPropertiesFileVariation) {
					String newLocation = location.substring(0, index) + File.separator+substring+ configurationPoint.getValue()+".properties";
					File f = new File(newLocation);
					if(f.exists() && !f.isDirectory()) { 
						request.setModelLocation(newLocation);
					}else {
						log.error("No configuration file at:" + newLocation);
						log.info("Skipping DoF " +configurationPoint.getValue());
						continue;
					}
						
				}else{
					request.setModelLocation(location.substring(0, index) + File.separator+ configurationPoint.getValue()+ File.separator);					
				}
				EntityMapping subResult = requestMetrics(request);
				addDoFSetting(configurationPoint.getKey().getIdentifier(), configurationPoint.getValue(), subResult);
				results.add(subResult);
			}
		}
		return results;
	}
	
	private void addDoFSetting(String identifier, String value, EntityMapping newRequest) {
		for (DoF dof : newRequest.getDoF())
			if (dof.getIdentifier().contains(identifier)) {
				dof.setEntityType("DOF");
				dof.setAlias(dof.getIdentifier());
				Probe p = DomainFactory.eINSTANCE.createProbe();
				dof.getProbes().add(p);
				StatType st = DomainFactory.eINSTANCE.createStatType();
				p.getStatTypes().add(st);
				StringResult result = DomainFactory.eINSTANCE.createStringResult();
				// instruction variation e.g. migrate
				if (value.contains(";;;;;"))
					result.setValue("Target deployment ID " + value.split(";;;;;")[1]);
				// value variation e.g. ADD 10
				else
					result.setValue(value);
				result.setValid(true);
				st.setResult(result);
			}
	}
	
//	private EntityMapping requestMetrics(EntityMapping request, HashMap<DoF, String> configuration,
//			Properties properties, String outputDirectory) {
//		final EntityMapping response = EcoreUtil.copy(request);
//		for (Entry<DoF, String> configurationPoint : configuration.entrySet()) {
////			addDoFSetting(configurationPoint.getKey().getIdentifier(), configurationPoint.getValue(), response);
////			models = ModelAdapter.adaptModel(models, configurationPoint.getKey(), configurationPoint.getValue());
//			
//		}
//		return requestMetrics(properties, response, outputDirectory);
//	}
	

	/* Modifies response to include the results of Kieker */
	private void processPerformanceResults(EntityMapping response, Trace kiekerResults, long firstTs, long lastTs) {

		Map<Resource, Trace> resources = kiekerResults.getResourceToTraceMap();
		Map<ResourceIdentifier, Resource> resourceIds = kiekerResults.getResourceIdentifierToResourceMap();
		Map<Component, Trace> services = kiekerResults.getComponentToTraceMap();

		// browse through each entity and its probes
		// and try to replace the probe with a result
		for (Entity ent : response.getEntities()) {
			if (ent.getIdentifier().equals("session") | ent.getAlias().equals("session")) {
				for (Probe dqlProbe : ent.getProbes()) {
					if(dqlProbe.getMetricName().equals(KiekerHelper.METRIC_THROUGHPUT)) {
						addSessionThroughput(ent, dqlProbe, firstTs, lastTs);
					}else {
						addSessionResponseTimes(ent, dqlProbe, firstTs);
					}
				}
			}else if (ent.getIdentifier().equals("system") | ent.getAlias().equals("system")) {
				addSystemUtilization(firstTs, resources, resourceIds, ent);
			
			} else {
				for (Probe dqlProbe : ent.getProbes()) {
					if(dqlProbe.getStatTypes().isEmpty()) {
						//dqlProbe.getStatTypes().add(new )
						log.error("NEED TO ADD DEFAULT STAT TYPE");
					}
					for (StatType st : dqlProbe.getStatTypes()) {
						if (st.getTypeName().equalsIgnoreCase("mean")){
							ResourceIdentifier dqlId = ResourceIdentifier.createResourceIdentifier("",
									ent.getIdentifier());
							Resource res = resourceIds.get(dqlId);
							if (res != null && resources.get(res) != null
									&& resources.get(res).getResources().size() > 0) {
								try {
									double value=0;
									int cnt = 0;
									for (ResourceRecord record : resources.get(res).getResources()) {
										value += record.getContent().get(ResourceContent.UTILIZATION) / 1000000000.0;
										cnt++;
									}
									createUtilization(ent, st, value/cnt);
								} catch (NumberFormatException exc) {
									System.out.println("Error "+exc.getStackTrace());
								}
							}
						}else if (KiekerHelper.METRIC_UTILIZATION.equalsIgnoreCase(dqlProbe.getMetricName())) {
							if (st.getTypeName().equalsIgnoreCase("DEFAULT") || st.getTypeName().equalsIgnoreCase("SAMPLE")) {
								ContinuousSeriesResult result = DomainFactory.eINSTANCE.createContinuousSeriesResult();
								result.setXAxisDescription("time in ms");
								result.setYAxisDescription("%"); // (between 0 and 1)");
								ResourceIdentifier dqlId = ResourceIdentifier.createResourceIdentifier("",
										ent.getIdentifier());
								Resource res = resourceIds.get(dqlId);
								if (res != null && resources.get(res) != null
										&& resources.get(res).getResources().size() > 0) {

									try {
										for (ResourceRecord record : resources.get(res).getResources()) {
											// System.out.println("ADDING for
											// "+ent.getIdentifier()+": Time"
											// +record.getTimestamp()+" / Util:
											// "+record.getContent().get(2));
											KiekerHelper.addToResult(result, record.getTimestamp(), record.getContent().get(ResourceContent.UTILIZATION), firstTs);
										}
										result.setValid(true);
									} catch (NumberFormatException exc) {
										System.out.println("Error "+exc.getStackTrace());
										result.setValid(false);
									}
									st.setResult(result);
								}
							} else {
								// do nothing
							}
						} else if (KiekerHelper.METRIC_RESPONSE_TIME.equalsIgnoreCase(dqlProbe.getMetricName())) {
							if (st.getTypeName().equalsIgnoreCase("DEFAULT") || st.getTypeName().equalsIgnoreCase("SAMPLE")) {
								ContinuousSeriesResult res = DomainFactory.eINSTANCE.createContinuousSeriesResult();
								res.setXAxisDescription("in s");	//seconds
								res.setYAxisDescription("arrival time in ms");
								if (services != null) {
									try {
										Component comp = new Component(
												KiekerHelper.getComponentString(ent.getIdentifier()));
										Trace tr = services.get(comp);
										if (tr != null && tr.getEvents().size() > 0) {
											for (EventRecord er : tr.getEvents()) {
												String dqlId = ent.getIdentifier();

												String kiekerId = er.getIdentifier().getComponentID();
												if (er.getIdentifier().isInterfaceIdentifier()) {
													kiekerId += "." + er.getIdentifier().getInterfaceID();
												}
												if (kiekerId.contains(dqlId)) {
													KiekerHelper.addToResult(res, er.getTimestamp(), er.getRuntime() / 1000000000.0, firstTs);
												}

											}
											res.setValid(true);
										}

									} catch (NumberFormatException exc) {
										log.warn("NumberFormatException occured");
										log.info(exc);
										res.setValid(false);
									}
									st.setResult(res);
								}
							}
							
						} else if (KiekerHelper.METRIC_THROUGHPUT.equalsIgnoreCase(dqlProbe.getMetricName())) {
							addServiceThroughput(firstTs, lastTs, services, ent, dqlProbe);

						}

					}
				}
			}
		}
		//DEBUG OUTPUT
		

		log.info("--------------------------------------------------------------------------------");
		log.info("\t\t\tListing results");
		log.info("--------------------------------------------------------------------------------");
		for(Entity entity:response.getEntities()) {
			log.info(entity.getIdentifier());
			for(Probe probe:entity.getProbes()) {
				log.info("\t"+probe.getMetricName() );
				//log.info("\t\t"+probe.getStatTypes());
				for (StatType statType:probe.getStatTypes()) {
					if(statType.getResult() instanceof ContinuousSeriesResult) {
						log.info("\t\t"+((ContinuousSeriesResult)statType.getResult()).getElements().size()+" values");				
					}else if (statType.getResult() instanceof DecimalResult) {
						//log.info("\t\t"+statType.getTypeName());
						log.info("\t\t"+((DecimalResult)statType.getResult()).getValue());				
					}
				
				}
			}

			

		}
	}

private void addSystemUtilization(long firstTs, Map<Resource, Trace> resources,
		Map<ResourceIdentifier, Resource> resourceIds, Entity ent) {
	ContinuousSeriesResult result = DomainFactory.eINSTANCE.createContinuousSeriesResult();
	result.setXAxisDescription("time in ms");
	result.setYAxisDescription("%"); // (between 0 and 1)");
	ResourceIdentifier systemidentifier = ResourceIdentifier.createResourceIdentifier("system", "system");
	
	List<ResourceRecord> system_utilization = new ArrayList<ResourceRecord>();
	
	log.info("Merging the results of "+ resourceIds.keySet().size()+ " cores to calculate utilization of system --- START");

	
	for (ResourceIdentifier resourceID : resourceIds.keySet()) {
		Resource resource = resourceIds.get(resourceID);
		List<ResourceRecord> cpu_utilization = resources.get(resource).getResources();
//		System.out.println("--------------------------------------------------------------------------------");
//		System.out.println("\t "+ resource.getName() +" has "+ cpu_utilization.size() + " entries");
//		System.out.println("--------------------------------------------------------------------------------");
		
		Comparator<? super ResourceRecord> resourceRecordComparator = new Comparator<ResourceRecord>() 
		{
			public int compare(ResourceRecord o1, ResourceRecord o2) {
				if(o2.getTimestamp() == o1.getTimestamp()) {
					return 0;
				}else if (o2.getTimestamp() - o1.getTimestamp() > 0){
					return -1;
				}else {
					return 1;
				}
			};
		};
		Collections.sort(cpu_utilization, resourceRecordComparator);
		
		for(int i=0; i<cpu_utilization.size(); i++) {
			if(i < system_utilization.size()) {
//				if(system_utilization.get(i).getTimestamp() == cpu_utilization.get(i).getTimestamp()) {
//				}else {
//					System.out.println("Timestamp missmatch "+(resource.getName()) +" "+ ((system_utilization.get(i).getTimestamp() - cpu_utilization.get(i).getTimestamp())/1000.0/1000.0) + " ms");
//				}
				double sum = system_utilization.get(i).getContent().get(ResourceContent.UTILIZATION) + 	cpu_utilization.get(i).getContent().get(ResourceContent.UTILIZATION);
				system_utilization.get(i).getContent().set(ResourceContent.UTILIZATION, sum);
			}else {
				system_utilization.add(new ResourceRecord(systemidentifier, cpu_utilization.get(i).getTimestamp(), cpu_utilization.get(i).getContent().clone()));
			}
		}
	}
	for(ResourceRecord record:system_utilization) {
		KiekerHelper.addToResult(result, record.getTimestamp(), record.getContent().get(ResourceContent.UTILIZATION) 
				/ (double)resourceIds.keySet().size() //
				, firstTs);
		
	}
	log.info("Merging the results of "+ resourceIds.keySet().size()+ " cores to calculate utilization of system --- FINISHED");
	
	ent.getProbes().get(0).setMetricName("utilization");
	ent.getProbes().get(0).getStatTypes().get(0).setTypeName("sample");
	ent.getProbes().get(0).getStatTypes().get(0).setResult(result);
}

	private void addServiceThroughput(long firstTs, long lastTs, Map<Component, Trace> services, Entity ent,
			Probe dqlProbe) {
		StatType statType = dqlProbe.getStatTypes().get(0);
		int cnt = 0;
		try {
			Component comp = new Component(
					KiekerHelper.getComponentString(ent.getIdentifier()));
			Trace tr = services.get(comp);
			if (tr != null && tr.getEvents().size() > 0) {
				for (EventRecord er : tr.getEvents()) {
					String dqlId = ent.getIdentifier();

					String kiekerId = er.getIdentifier().getComponentID();
					if (er.getIdentifier().isInterfaceIdentifier()) {
						kiekerId += "." + er.getIdentifier().getInterfaceID();
					}
					if (kiekerId.contains(dqlId)) {
						cnt++;
					}

				}
			}

		} catch (NumberFormatException exc) {
			log.error(exc);
		}

		createThroughput(firstTs, lastTs, ent, statType, cnt);
	}

	private void createThroughput(long firstTs, long lastTs, Entity ent, StatType statType, int cnt) {
		DecimalResult result = DomainFactory.eINSTANCE.createDecimalResult();
		double interval = ((lastTs - firstTs)/1000.0/1000.0/1000.0);	//in seconds, micro, nano, milli
		long value = (long) (cnt / interval);
		log.info("requests "+ent.getIdentifier()+": "+cnt);
		log.info("interval "+ interval + "       ["+firstTs+" : "+lastTs+"]");
		log.info("througput" + new BigDecimal(value));
		result.setValue(new BigDecimal(value));
		result.setValid(true);
		statType.setResult(result);
	}
	
	private void createUtilization(Entity ent, StatType statType, double value) {
		DecimalResult result = DomainFactory.eINSTANCE.createDecimalResult();
		result.setValue(new BigDecimal(value));
		result.setValid(true);
		statType.setResult(result);
	}

	private void addSessionResponseTimes(Entity entity, Probe dqlProbe,
			long firstTs) {
		StatType statType = dqlProbe.getStatTypes().get(0);

		ContinuousSeriesResult result = DomainFactory.eINSTANCE.createContinuousSeriesResult();
		result.setXAxisDescription("response time in ms");
		result.setYAxisDescription("arrival time in ms");
//		String abc = "session"; // ent.getIdentifier();
//		ResourceIdentifier dqlId = ResourceIdentifier.createResourceIdentifier("", abc);
//		Resource res = new Resource("session");// resourceIds.get(dqlId);
		// resourceIds.put(dqlId, res);
		for (Session session : SessionFilter.sessionTimes) {
			KiekerHelper.addToResult(result, session.start, session.duration / 1000000.0,
					firstTs);
		}
		statType.setResult(result);
	}
	
	private void addSessionThroughput(Entity entity, Probe dqlProbe, long firstTs, long lastTs) {
		StatType statType = dqlProbe.getStatTypes().get(0);
		createThroughput(firstTs, lastTs, entity, statType, SessionFilter.sessionTimes.size());
	}
	

	public void reset() {
		// Nothing to do here
	}

	@Override
	public void addQueryProgressListener(QueryProgressListener arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public String getConstraint() {
		return null;
	}

	@Override
	public String getDoFConstraint() {
		return null;
	}

	@Override
	public String getDoFExplorationStrategy() {
		return null;
	}

	@Override
	public Properties getDoFExplorationStrategyProperties() {
		return null;
	}

	@Override
	public List<String> getSupportedConstraintNames() {
		return null;
	}

	@Override
	public List<String> getSupportedDoFConstraintNames() {
		return null;
	}

	@Override
	public List<String> getSupportedExplorationStrategyNames() {
		return null;
	}

	@Override
	public List<String> getSupportedTimeUnits() {
		return null;
	}

	@Override
	public void setConstraint(final String constraint) {
	}

	@Override
	public void setDoFConstraint(final String doFConstraint) {
	}

	@Override
	public void setDoFExplorationStrategy(final String strategy, final Properties properties) {
	}

	@Override
	public void setObservationPeriod(final Date start, final Date end, final long sampleSize) {
	}

	@Override
	public void setObservationPeriod(final long amount, final String unit, final long sampleSize) {
	}

	@Override
	public void setObservationReference(final String type, final String identifier, final long sampleSize) {
	}

	@Override
	public boolean supportsConstraintRequests() {
		return false;
	}

	@Override
	public boolean supportsDoFRequests() {
		return true;
	}

	@Override
	public boolean supportsMetricRequests() {
		return true;
	}

	@Override
	public boolean supportsObservationPeriods() {
		return false;
	}
}