package tools.descartes.dql.connector.kieker.util;

import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;

import kieker.common.configuration.Configuration;
import tools.descartes.dql.connector.kieker.structures.Trace;

public class KiekerAnalysisController {
	private final Configuration config;
	public static final String CONFIG_LOG4J_FILE_PATH = "LOG4J_FILE_PATH";
	private Trace trace;
	private boolean ignoreHost = false;

	public KiekerAnalysisController(Configuration config, boolean ignoreHost) {
		this.config=config;
		this.ignoreHost = ignoreHost;
	}

	public KiekerAnalysisController(String configFilePath) {
		Configuration config = new Configuration();
		BufferedInputStream stream;
		try {
			stream = new BufferedInputStream(new FileInputStream(configFilePath));
			config.load(stream);
			stream.close();
		} catch (IOException e) {
			System.out.println("no config for this trace; default is used");
			System.out.println(e.getMessage());
		}
		this.config = config;

	}

	public Trace getTrace() {
		return trace;
	}

	public void fillTrace() {
		Trace t = new Trace();
		ArrayList<String> traceDirectories = new ArrayList<String>();

		for (String propertyKeyString : config.stringPropertyNames()) {
			String propertyKey = propertyKeyString.trim();
			System.out.println("PROPKEY: " + propertyKey);
			if (propertyKey.startsWith("COMPONENT_") && propertyKey.length() >= 10) {
				String componentName = propertyKey.substring(10).trim();
				if (componentName.contains(".")) {
					try {
						componentName = componentName.split("\\.")[0];
					} catch (Exception e) {
						throw new IllegalArgumentException(
								"invalid component declaration format in configuration file");
					}
				}
				t.setComponentDefinition(componentName, config.getStringProperty(propertyKeyString).trim());
			} else if (propertyKey.startsWith("RESOURCE_") && propertyKey.length() >= 9) {
				String resourceName = propertyKey.substring(9).trim();
				if (resourceName.contains(".")) {
					try {
						resourceName = resourceName.split("\\.")[0];
					} catch (Exception e) {
						throw new IllegalArgumentException("invalid resource declaration format in configuration file");
					}
				}
				t.setResourceDefinition(resourceName, config.getStringProperty(propertyKeyString).trim());
			} else if (propertyKey.startsWith("TRACE_") && propertyKey.length() >= 6) {
				// String traceName = propertyKey.substring(6).trim();
				String tracePath = config.getStringProperty(propertyKeyString).trim();
				traceDirectories.add(tracePath);
			}
		}

		KiekerTraceLoader loader = new KiekerTraceLoader(t);
		t =loader.loadKiekerTraceFromDirectories(traceDirectories.toArray(new String[traceDirectories.size()]));
		this.trace = t.initializeMaps(ignoreHost);

	}

	public boolean evaluate() {
		return false;
	}

	/*
	 * public static void main(String[] args) { public Trace run() { //
	 * BasicConfigurator.configure(); //Configuration config = new
	 * Configuration();
	 * //config.setProperty(KiekerController.CONFIG_LOG4J_FILE_PATH,
	 * modelLocation);
	 * 
	 * KiekerController c = new KiekerController(modelLocation +
	 * "//controller.cfg"); c.fillTrace();
	 * 
	 * 
	 * Trace trac = c.getTrace(); System.out.println("Printing result now"); //
	 * System.out.println(t.getComponentIdentifierToComponentMap());
	 * System.out.println(trac.getResourceIdentifierToResourceMap()); //
	 * t.printEventsPerRequest(); // t.printRequestsPerComponent();
	 * trac.printResources(); return trac;
	 * 
	 * /* for (Entry<String, Request> entry : t.getIdToRequestMap().entrySet())
	 * { // services.add(WorkloadBuilder.newService(entry.getKey()));
	 * List<EventRecord> records = entry.getValue().getEvents(); for
	 * (EventRecord eventRecord : records) { System.out.println("RECORD: " +
	 * eventRecord.toString()); eventRecord.getTimestamp();
	 * eventRecord.getRuntime(); } }
	 */

		// System.out.println("==========================");
		// System.out.println("##### ORIGINAL INPUT #####");
		// System.out.println("===========CLASS==========");
		// System.out.println(ClassMap.getInstance());
		// System.out.println("=========INTERFACE========");
		// System.out.println(InterfaceMap.getInstance());
		// System.out.println("==========RESOURCE========");
		// System.out.println(ResourceMap.getInstance());
		// System.out.println("============CPU===========");
		// System.out.println(CpuMap.getInstance());
		// System.out.println("==========MEM/SWAP========");
		// System.out.println(MemSwapMap.getInstance());
		//
		// System.out.println(t.getComponentToTraceMap());

		//c.exportToLibrede();
	// }
}
